/*
 *  KeyboardController.java
 *  Joris van Rantwijk
 */

package retro;

/**
 * Simplified simulation of the 8255 XT keyboard controller
 * and Programmable Peripheral Interface.
 *
 * We don't even try to be precise since the AT keyboard controller
 * is very different and most software works with both controllers anyway.
 * The implemented behaviour falls somewhere between XT and AT. 
 * <p>
 * We assume 8255 command byte 0x99:
 * <ul>
 *  <li> Port A in mode 0 input for keyboard data
 *  <li> Port B in mode 0 output for system control
 *  <li> Port C in mode 0 input for system status
 * </ul>
 */
public class KeyboardController implements IOPortHandler
{

    private Scheduler sched;
    private InterruptRequest irq;
    private int switchS1;
    private int switchS2;
	
    private int port61;
    private String keyBuf;
    private int lastKeyCode;
    boolean keyShiftPending;
    private Scheduler.Task keyShiftTask;


    /** Construct and initialize keyboard controller simulation. */
    public KeyboardController(Scheduler sched, InterruptRequest irq)
    {
        this.sched = sched;
        this.irq = irq;
        keyBuf = new String();
        keyShiftPending = false;
        keyShiftTask = new Scheduler.Task() {
            public void run() {
                // raise interrupt again for the next pending byte
                keyShiftPending = false;
                keyBuf = keyBuf.substring(1);
                if (keyBuf.length() > 0 && KeyboardController.this.irq != null)
                    KeyboardController.this.irq.raise(true);
            }
        };
    }


    /** Set configuration switch data to be reported by PPI. */
    public void setSwitchData(int S1, int S2)
    {
        switchS1 = S1;
        switchS2 = S2;
    }


    /** Store a scancode byte in the buffer. */
    public void putKeyData(int v)
    {
        if (keyBuf.length() == 0 && irq != null)
            irq.raise(true);
        keyBuf = keyBuf + ((char)v);
    }


    /** Store scancode bytes in the buffer. */
    public void putKeyData(byte[] b)
    {
        if (keyBuf.length() == 0 && b.length > 0 && irq != null)
            irq.raise(true);
        StringBuffer sb = new StringBuffer(keyBuf);
        for (int i = 0; i < b.length; i++)
            sb.append((char)(b[i] & 0x00ff));
        keyBuf = sb.toString();
    }


    /** Get a scancode byte from the buffer. */
    private final int getKeyData()
    {
        // release interrupt
        if (irq != null)
            irq.raise(false);
        // if the buffer is empty, we just return the most recent byte 
        if (keyBuf.length() > 0) {
            // read byte from buffer
            lastKeyCode = keyBuf.charAt(0);
            // wait .5 msec before going to the next byte
            if (!keyShiftPending) {
                keyShiftPending = true;
                sched.runTaskAfter(keyShiftTask, 500000);
            }
        }
        // return scancode byte
        return lastKeyCode;
    }


    /** Get status byte for system control port. */
    private int getStatusByte()
    {
        // bit 0: timer 2 clock gate to speaker status
        // bit 1: speaker data status
        // bit 2: NMI parity check status
        // bit 3: NMI I/O check status
        // bit 4: toggles with each refresh count (TODO or is it timer 1 output??)
        // bit 5: timer 2 output
        // bit 6: I/O channel parity error occurred
        // bit 7: RAM parity error occurred
        int v = (port61 & 0x03) | ((port61 >> 2) & 0x0c);
        // TODO : get timer channel 2 output
        // TODO : get refresh count
        return v;
    }


    /** Handle write request from Cpu. */
    public void outb(int v, int port)
    {
        if ((port & 3) == 1) {
            // Write to port 0x61 (system control port)
            // bit 0: timer 2 gate to speaker
            // bit 1: speaker data
            // bit 3: read low(0) or high(1) nibble of S2 switches
            // bit 4: NMI RAM parity check disable
            // bit 5: NMI I/O check disable
            // bit 6: enable(1) or disable(0) keyboard clock ??
            // bit 7: pulse 1 to reset keyboard and IRQ1
            // TODO : signal speaker status
            port61 = v;
            //System.err.println("Keyboard: OUTB " + Misc.byteToHex(v) + " -> " + Misc.wordToHex(port));
        } else {
            System.err.println("Keyboard: OUTB " + Misc.byteToHex(v) + " -> " + Misc.wordToHex(port) + " not implemented");
        }
    }


    /** Handle read request from Cpu. */
    public int inb(int port)
    {
        switch (port & 3) {
          case 0: // port 0x60 (PPI port A)
            return ((port61 & 0x80) == 0) ? getKeyData() : switchS1;
          case 1: // port 0x61 (PPI port B)
            // huh? port B is supposed to be output mode!
            // apparently this is used on AT and maybe also XT ??
            //System.err.println("Keyboard: INB " + Misc.byteToHex(port));
            return getStatusByte();
           case 2: // port 0x62 (PPI port C)
            return ((port61 & 0x08) == 0) ?
                   ((getStatusByte() & 0xf0) | (switchS2 & 0x0f)) :
                   ((getStatusByte() & 0xf0) | ((switchS2 >> 4) & 0x0f));
          default: // port 0x63 (PPI command byte)
              return 0x99;
        }
    }

}

/* end */
