/*
 *  KeyboardExt.java
 *  Joris van Rantwijk
 */

/*
 * TODO : implement E0 escape scheme
 * TODO : change this from a passive mapping into a filter which feeds the keyboard controller
 */

package retro;
import java.awt.event.KeyEvent;

/**
 * KeyboardExt simulates a 101-key extended PC keyboard.
 *
 * A KeyboardExt object accepts key up/down events using Java VK_xxx-style
 * key codes, and sends keyboard scan codes to the keyboard controller.
 * <p>
 * This object produces XT scan codes for the 83 keys of the original
 * PC keyboard, and e0-style extended scan codes for the new keys of
 * the 101-key keyboard. This is commonly known as scan code set 1, and
 * corresponds to the default mode of operation of physical 101-key
 * extended keyboards.
 * <p>
 * Inferring the correct escape sequence for extended keys requires
 * knowledge of the state of modifier (Shift, Ctrl, Alt) keys and NumLock.
 * This simulation keeps track of the state of modifier keys by tracking
 * the corresponding key up/down events. In addition, the modifier mask
 * from the Java event object may be used to synchronize these states
 * (for example, after losing key events that occurred while the simulator
 * did not have keyboard focus).
 * <p>
 * Tracking the NumLock state on a PC/XT is problematic because there
 * is no feedback from the software; thus the keyboard and software can
 * get out of sync. Physical extended XT keyboards also have this problem.
 * For key mapping, the NumLock state may be obtained from the actual
 * NumLock state of the host system; this effectively shifts the
 * synchronization problem to the user of the emulator, who will presumably
 * find a way to deal with it.
 */
public class KeyboardExt
{

    /** Map Java key code to PC scan code and extra flags. */
    private int[] keytbl;

    /** Must send E0 escape with this scancode */
    private static final int KEY_EXTEND = 0x100;

    /** Need negative shift/numlock state */
    private static final int KEY_NONUM = 0x200;

    /** Need positive shift state */
    private static final int KEY_ISSHIFT = 0x400;

    /** Could be numpad or edit block depending on key location */
    private static final int KEY_EDIT = 0x800;

    /** Virtual state of shift keys induced by previous scan codes */
    private boolean virtualLeftShift, virtualRightShift;

    /** True if we believe that the right shift key is currently down */
    private boolean rightShiftState;


    /** Create and initialize key mapper. */
    public KeyboardExt()
    {
        virtualLeftShift = false;
        virtualRightShift = false;
        rightShiftState = false;
        keytbl = new int[256];
        keytbl[KeyEvent.VK_ESCAPE] = 1;
        keytbl[KeyEvent.VK_1] = 2;
        keytbl[KeyEvent.VK_2] = 3;
        keytbl[KeyEvent.VK_3] = 4;
        keytbl[KeyEvent.VK_4] = 5;
        keytbl[KeyEvent.VK_5] = 6;
        keytbl[KeyEvent.VK_6] = 7;
        keytbl[KeyEvent.VK_7] = 8;
        keytbl[KeyEvent.VK_8] = 9;
       	keytbl[KeyEvent.VK_9] = 10;
        keytbl[KeyEvent.VK_0] = 11;
        keytbl[KeyEvent.VK_MINUS] = 12;
        keytbl[KeyEvent.VK_EQUALS] = 13;
        keytbl[KeyEvent.VK_BACK_SPACE] = 14;
        keytbl[KeyEvent.VK_TAB] = 15;
        keytbl[KeyEvent.VK_Q] = 16;
        keytbl[KeyEvent.VK_W] = 17;
        keytbl[KeyEvent.VK_E] = 18;
        keytbl[KeyEvent.VK_R] = 19;
        keytbl[KeyEvent.VK_T] = 20;
        keytbl[KeyEvent.VK_Y] = 21;
        keytbl[KeyEvent.VK_U] = 22;
        keytbl[KeyEvent.VK_I] = 23;
        keytbl[KeyEvent.VK_O] = 24;
        keytbl[KeyEvent.VK_P] = 25;
        keytbl[KeyEvent.VK_OPEN_BRACKET] = 26;
        keytbl[KeyEvent.VK_CLOSE_BRACKET] = 27;
        keytbl[KeyEvent.VK_ENTER] = 28;
        keytbl[KeyEvent.VK_CONTROL] = 29;
        keytbl[KeyEvent.VK_A] = 30;
        keytbl[KeyEvent.VK_S] = 31;
        keytbl[KeyEvent.VK_D] = 32;
        keytbl[KeyEvent.VK_F] = 33;
        keytbl[KeyEvent.VK_G] = 34;
        keytbl[KeyEvent.VK_H] = 35;
        keytbl[KeyEvent.VK_J] = 36;
        keytbl[KeyEvent.VK_K] = 37;
        keytbl[KeyEvent.VK_L] = 38;
        keytbl[KeyEvent.VK_SEMICOLON] = 39;
        keytbl[KeyEvent.VK_QUOTE] = 40;
        keytbl[KeyEvent.VK_BACK_QUOTE] = 41;
        keytbl[KeyEvent.VK_SHIFT] = 42; // left shift
        keytbl[KeyEvent.VK_BACK_SLASH] = 43;
        keytbl[KeyEvent.VK_Z] = 44;
        keytbl[KeyEvent.VK_X] = 45;
        keytbl[KeyEvent.VK_C] = 46;
        keytbl[KeyEvent.VK_V] = 47;
        keytbl[KeyEvent.VK_B] = 48;
        keytbl[KeyEvent.VK_N] = 49;
        keytbl[KeyEvent.VK_M] = 50;
        keytbl[KeyEvent.VK_COMMA] = 51;
        keytbl[KeyEvent.VK_PERIOD] = 52;
        keytbl[KeyEvent.VK_SLASH] = 53;
        keytbl[KeyEvent.VK_DIVIDE] = 53 | KEY_EXTEND;
        keytbl[KeyEvent.VK_MULTIPLY] = 55;
        keytbl[KeyEvent.VK_ALT] = 56;
        keytbl[KeyEvent.VK_ALT_GRAPH] = 56 | KEY_EXTEND;
        keytbl[KeyEvent.VK_SPACE] = 57;
        keytbl[KeyEvent.VK_CAPS_LOCK] = 58;
        keytbl[KeyEvent.VK_F1] = 59;
        keytbl[KeyEvent.VK_F2] = 60;
        keytbl[KeyEvent.VK_F3] = 61;
        keytbl[KeyEvent.VK_F4] = 62;
        keytbl[KeyEvent.VK_F5] = 63;
        keytbl[KeyEvent.VK_F6] = 64;
        keytbl[KeyEvent.VK_F7] = 65;
        keytbl[KeyEvent.VK_F8] = 66;
        keytbl[KeyEvent.VK_F9] = 67;
        keytbl[KeyEvent.VK_F10] = 68;
        keytbl[KeyEvent.VK_NUM_LOCK] = 69;
        keytbl[KeyEvent.VK_SCROLL_LOCK] = 70;
        keytbl[KeyEvent.VK_NUMPAD7] = 71;
        keytbl[KeyEvent.VK_HOME] = 71 | KEY_EDIT;
        keytbl[KeyEvent.VK_NUMPAD8] = 72;
        keytbl[KeyEvent.VK_UP] = 72 | KEY_EXTEND | KEY_NONUM;
        keytbl[KeyEvent.VK_KP_UP] = 72;
        keytbl[KeyEvent.VK_NUMPAD9] = 73;
        keytbl[KeyEvent.VK_PAGE_UP] = 73 | KEY_EDIT;
        keytbl[KeyEvent.VK_SUBTRACT] = 74;
        keytbl[KeyEvent.VK_NUMPAD4] = 75;
        keytbl[KeyEvent.VK_LEFT] = 75 | KEY_EXTEND | KEY_NONUM;
        keytbl[KeyEvent.VK_KP_LEFT] = 75;
        keytbl[KeyEvent.VK_NUMPAD5] = 76;
        keytbl[KeyEvent.VK_NUMPAD6] = 77;
        keytbl[KeyEvent.VK_RIGHT] = 77 | KEY_EXTEND | KEY_NONUM;
        keytbl[KeyEvent.VK_KP_RIGHT] = 77;
        keytbl[KeyEvent.VK_ADD] = 78;
        keytbl[KeyEvent.VK_NUMPAD1] = 79;
        keytbl[KeyEvent.VK_END] = 79 | KEY_EDIT;
        keytbl[KeyEvent.VK_NUMPAD2] = 80;
        keytbl[KeyEvent.VK_DOWN] = 80 | KEY_EXTEND | KEY_NONUM;
        keytbl[KeyEvent.VK_KP_DOWN] = 80;
        keytbl[KeyEvent.VK_NUMPAD3] = 81;
        keytbl[KeyEvent.VK_PAGE_DOWN] = 81 | KEY_EDIT;
        keytbl[KeyEvent.VK_NUMPAD0] = 82;
        keytbl[KeyEvent.VK_INSERT] = 82 | KEY_EDIT;
        keytbl[KeyEvent.VK_DECIMAL] = 83;
        keytbl[KeyEvent.VK_DELETE] = 83 | KEY_EDIT;
        keytbl[KeyEvent.VK_PRINTSCREEN] = 84; // sysreq // NO, THIS IS MORE COMPLICATED
        keytbl[KeyEvent.VK_F11] = 86;
        keytbl[KeyEvent.VK_F12] = 87;
        // TODO : VK_PRINTSCREEN should be done differently
        // TODO : VK_PAUSE
    }


    /** Translate a Java keyPressed event into a make scan code sequence. */
    public byte[] getMakeCode(int keyCode, int modifiers, int keyLocation,
                              boolean numlock)
    {
        // Get key info from table
        int keyinfo = 0;
        if (keyCode < keytbl.length)
            keyinfo = keytbl[keyCode];
        if (keyinfo == 0) {
            System.err.println("getMakeCode: unknown keycode " + keyCode + "=" + KeyEvent.getKeyText(keyCode));
            return null;
        }
        // Distinguish between numpad and non-nunpam edit keys
        if ((keyinfo & KEY_EDIT) != 0) {
            keyinfo = (keyinfo & (~KEY_EDIT));
            if (keyLocation != KeyEvent.KEY_LOCATION_NUMPAD)
                keyinfo |= KEY_EXTEND | KEY_NONUM;
        }
        // Distinguish between left and right ctrl, alt
        if (keyLocation == KeyEvent.KEY_LOCATION_RIGHT &&
            (keyCode == KeyEvent.VK_CONTROL || keyCode == KeyEvent.VK_ALT))
            keyinfo |= KEY_EXTEND;
        byte[] b = new byte[1];
        // b[0] = (byte)scancode;
        return b;
    }


	// Translate a Java keyReleased event into a break scan code sequence
	public byte[] getBreakCode(int keyCode, int modifiers, int keyLocation)
	{
		int scancode = 0;
		if (keyCode < keytbl.length)
			scancode = keytbl[keyCode];
		if (scancode == 0) {
			System.err.println("getBreakCode: unknown keycode " + keyCode + "=" + KeyEvent.getKeyText(keyCode));
			return null;
		}
		byte[] b = new byte[1];
		b[0] = (byte) (scancode | 0x80);
		return b;
	}


	// Translate a Java keyPressed event into a repeat scan code sequence
	public byte[] getRepeatCode(int keyCode, int modifiers, int keyLocation)
	{
		// TODO : how is keyboard repeat handled on XT ??
		int scancode = 0;
		if (keyCode < keytbl.length)
			scancode = keytbl[keyCode];
		if (scancode == 0)
			return null;
		switch (scancode) {
		  case 29: // control
		  case 42: // left shift
		  case 54: // right shift
		  case 56: // alt
		  case 58: // caps lock
		  case 69: // num lock
		  case 70: // scroll lock
		  // TODO : probably add Insert
			// non-repeating keys
			return null;
		}
		byte[] b = new byte[1];
		b[0] = (byte)scancode;
		return b;
	}

}

/* end */
